from datetime import date
from pathlib import Path

import shell_interface as sh
import storage_device_managers as sdm

from . import config_parser as cp


def prepare_device_for_butterbackend(device: Path) -> cp.BtrFSRsyncConfig:
    password_cmd = sdm.generate_passcmd()
    backup_repository_folder = "ButterBackupRepository"
    volume_uuid = sdm.encrypt_device(device, password_cmd)
    compression = sdm.ValidCompressions.ZSTD
    user = sh.get_user()
    group = sh.get_group(user)
    with sdm.decrypted_device(device, password_cmd) as decrypted:
        sdm.mkfs_btrfs(decrypted)
        with sdm.mounted_device(decrypted) as mounted:
            backup_repository = mounted / backup_repository_folder
            mkdir_cmd: sh.StrPathList = ["sudo", "mkdir", backup_repository]
            sh.run_cmd(cmd=mkdir_cmd)

            initial_subvol = backup_repository / date.today().strftime(
                cp.BtrFSRsyncConfig.SubvolTimestampFmt
            )
            subvol_cmd: sh.StrPathList = [
                "sudo",
                "btrfs",
                "subvolume",
                "create",
                initial_subvol,
            ]
            sh.run_cmd(cmd=subvol_cmd)
            sdm.chown(mounted, user, group, recursive=True)

    config = cp.BtrFSRsyncConfig(
        BackupRepositoryFolder=backup_repository_folder,
        Compression=compression,
        DevicePassCmd=password_cmd,
        Files=set(),
        FilesDest="Einzeldateien",
        Folders={},
        UUID=volume_uuid,
    )
    return config


def prepare_device_for_resticbackend(device: Path) -> cp.ResticConfig:
    device_passcmd = sdm.generate_passcmd()
    repository_passcmd = sdm.generate_passcmd()
    backup_repository_folder = "ResticBackupRepository"
    compression = None  # Restic encrypts and encrypted data are incompressible
    volume_uuid = sdm.encrypt_device(device, device_passcmd)
    user = sh.get_user()
    group = sh.get_group(user)
    with sdm.decrypted_device(device, device_passcmd) as decrypted:
        sdm.mkfs_btrfs(decrypted)
        with sdm.mounted_device(decrypted) as mounted:
            backup_repo = mounted / backup_repository_folder
            mkdir_repo: sh.StrPathList = ["sudo", "mkdir", backup_repo]
            restic_init: sh.StrPathList = [
                "sudo",
                "restic",
                "init",
                "-r",
                backup_repo,
            ]
            sh.run_cmd(cmd=mkdir_repo)
            sh.pipe_pass_cmd_to_real_cmd(repository_passcmd, restic_init)
            sdm.chown(mounted, user, group, recursive=True)
    config = cp.ResticConfig(
        BackupRepositoryFolder=backup_repository_folder,
        Compression=compression,
        DevicePassCmd=device_passcmd,
        FilesAndFolders=set(),
        RepositoryPassCmd=repository_passcmd,
        UUID=volume_uuid,
    )
    return config



import abc
import datetime as dt
from dataclasses import dataclass
from pathlib import Path
from typing import overload

import shell_interface as sh
from loguru import logger

from . import config_parser as cp


class BackupBackend(abc.ABC):
    @abc.abstractmethod
    def do_backup(self, mount_dir: Path) -> None: ...

    @overload
    @staticmethod
    def from_config(config: cp.BtrFSRsyncConfig) -> "BtrFSRsyncBackend": ...

    @overload
    @staticmethod
    def from_config(config: cp.ResticConfig) -> "ResticBackend": ...

    @staticmethod
    def from_config(
        config: cp.BtrFSRsyncConfig | cp.ResticConfig,
    ) -> "BtrFSRsyncBackend | ResticBackend":
        # Getestet durch Tests der Backuplogik
        if isinstance(config, cp.BtrFSRsyncConfig):
            return BtrFSRsyncBackend(config=config)
        return ResticBackend(config=config)


@dataclass(frozen=True)
class BtrFSRsyncBackend(BackupBackend):
    config: cp.BtrFSRsyncConfig

    def do_backup(self, mount_dir: Path) -> None:
        logger.info(f"Beginne mit BtrFS-Backup für Speichermedium {self.config.UUID}.")
        backup_repository = mount_dir / self.config.BackupRepositoryFolder
        src_snapshot = self.get_source_snapshot(backup_repository)
        backup_root = self.snapshot(
            src=src_snapshot, backup_repository=backup_repository
        )
        for src, dest_name in self.config.Folders.items():
            dest = backup_root / dest_name
            self.rsync_folder(src, dest, self.config.ExcludePatternsFile)

        files_dest = backup_root / self.config.FilesDest
        if files_dest.is_file():
            files_dest.unlink()
        files_dest.mkdir(parents=True, exist_ok=True)
        for src in self.config.Files:
            self.rsync_file(src, files_dest)

    @staticmethod
    def get_source_snapshot(root: Path) -> Path:
        return max(root.glob("202?-*"))

    def snapshot(self, *, src: Path, backup_repository: Path) -> Path:
        timestamp = dt.datetime.now()
        date_fmt = self.config.SubvolTimestampFmt
        while True:
            # In order to get rid of sleeps in the tests, snapshot name
            # collisions must be handled gracefully. Incrementing the timestamp
            # by a few seconds seems to be acceptable both in test and real
            # world scenarios.
            # There is still the possibility of race conditions, where another
            # process would create the target folder before a snapshot can be
            # created. This possibility is believed to be negligible, so
            # nothing is done to prevent it.
            backup_root = backup_repository / timestamp.strftime(date_fmt)
            if backup_root.exists():
                timestamp += dt.timedelta(seconds=1)
            else:
                break
        cmd: sh.StrPathList = [
            "sudo",
            "btrfs",
            "subvolume",
            "snapshot",
            src,
            backup_root,
        ]
        sh.run_cmd(cmd=cmd)
        return backup_root

    @staticmethod
    def rsync_file(src: Path, dest: Path) -> None:
        cmd: sh.StrPathList = ["sudo", "rsync", "-ax", "--inplace", src, dest]
        sh.run_cmd(cmd=cmd)

    @staticmethod
    def rsync_folder(
        src: Path, dest: Path, maybe_exclude_patterns: Path | None
    ) -> None:
        cmd: sh.StrPathList = [
            "sudo",
            "rsync",
            "-ax",
            "--delete",
            "--delete-excluded",
            "--inplace",
        ]
        if maybe_exclude_patterns is not None:
            cmd.extend(["--exclude-from", maybe_exclude_patterns])
        cmd.extend([f"{src}/", dest])
        sh.run_cmd(cmd=cmd)


@dataclass(frozen=True)
class ResticBackend(BackupBackend):
    config: cp.ResticConfig

    def do_backup(self, mount_dir: Path) -> None:
        logger.info(f"Beginne mit Restic-Backup für Speichermedium {self.config.UUID}.")
        backup_repository = mount_dir / self.config.BackupRepositoryFolder
        self.copy_files(backup_repository)
        self.adapt_ownership(backup_repository)

    @staticmethod
    def adapt_ownership(backup_repository: Path) -> None:
        user = sh.get_user()
        group = sh.get_group(user)
        chown_cmd: sh.StrPathList = [
            "sudo",
            "chown",
            "-R",
            f"{user}:{group}",
            backup_repository,
        ]
        sh.run_cmd(cmd=chown_cmd)

    def copy_files(self, backup_repository: Path) -> None:
        restic_cmd: sh.StrPathList = [
            "sudo",
            "restic",
            "backup",
            "--one-file-system",
            "--repo",
            backup_repository,
        ]
        if self.config.ExcludePatternsFile is not None:
            restic_cmd.extend(["--exclude-file", self.config.ExcludePatternsFile])
        restic_cmd.extend(list(self.config.FilesAndFolders))
        sh.pipe_pass_cmd_to_real_cmd(self.config.RepositoryPassCmd, restic_cmd)



import json
import sys
import uuid
from collections import Counter
from pathlib import Path
from typing import ClassVar

from pydantic import (
    BaseModel,
    ConfigDict,
    DirectoryPath,
    FilePath,
    TypeAdapter,
    field_validator,
    model_validator,
)
from storage_device_managers import ValidCompressions

FoldersT = dict[DirectoryPath, str]


def path_aware_btrfs_json_decoding(folders: FoldersT) -> str:
    as_dict = {str(key): val for key, val in folders.items()}
    return json.dumps(as_dict)


def path_aware_restic_json_decoding(
    files_and_folders: set[FilePath | DirectoryPath],
) -> str:
    as_dict = {str(cur) for cur in files_and_folders}
    return json.dumps(as_dict)


class BaseConfig(BaseModel):
    BackupRepositoryFolder: str
    DevicePassCmd: str
    ExcludePatternsFile: FilePath | None = None
    UUID: uuid.UUID
    Compression: ValidCompressions | None = None

    @field_validator("ExcludePatternsFile", mode="before")
    def expand_tilde_in_exclude_patterns_file_name(
        cls, maybe_exclude_patterns
    ) -> str | None:
        if maybe_exclude_patterns is None:
            return None
        return str(Path(maybe_exclude_patterns).expanduser())

    def device(self) -> Path:
        return Path(f"/dev/disk/by-uuid/{self.UUID}")

    def map_name(self) -> Path:
        return Path(f"/dev/mapper/{self.UUID}")


class BtrFSRsyncConfig(BaseConfig):
    model_config = ConfigDict(extra="forbid", frozen=True)
    Files: set[FilePath]
    FilesDest: str
    Folders: FoldersT
    SubvolTimestampFmt: ClassVar[str] = "%F_%H:%M:%S"

    @field_validator("Files")
    def source_file_names_must_be_unique(cls, files):
        file_names = Counter(f.name for f in files)
        cls.raise_with_message_upon_duplicate(file_names, ("Dateinamen", "Dateinamen"))
        return files

    @field_validator("Folders")
    def folder_destinations_must_be_unique(cls, folders: FoldersT) -> FoldersT:
        destinations = Counter(folders.values())
        cls.raise_with_message_upon_duplicate(
            destinations, ("Zielverzeichnissen", "Ziele")
        )
        return folders

    @field_validator("Files", mode="before")
    def expand_tilde_in_file_sources(cls, files) -> list[str]:
        new = [str(Path(cur).expanduser()) for cur in files]
        return new

    @field_validator("Folders", mode="before")
    def expand_tilde_in_folder_sources(cls, folders) -> dict[str, str]:
        new = {str(Path(src).expanduser()): dest for src, dest in folders.items()}
        return new

    @model_validator(mode="after")
    def files_dest_is_no_folder_dest(self):
        files_dest = self.FilesDest
        destinations = self.Folders.values()
        if files_dest in destinations:
            raise ValueError(
                f"Zielverzeichnis {files_dest} ist gleichzeitig Ziel für Ordner und Einzeldateien."
            )
        return self

    @staticmethod
    def raise_with_message_upon_duplicate(
        counts: Counter[Path] | Counter[str], token: tuple[str, str]
    ) -> None:
        if all(val == 1 for val in counts.values()):
            return
        errmsg_begin = (
            f"Duplikate in {token[0]} entdeckt. Folgende {token[1]} kommen doppelt vor:"
        )
        errmsg_body = " ".join(
            str(elem) for (elem, count) in counts.items() if count > 1
        )
        raise ValueError(f"{errmsg_begin} {errmsg_body}")


class ResticConfig(BaseConfig):
    model_config = ConfigDict(extra="forbid", frozen=True)
    FilesAndFolders: set[FilePath | DirectoryPath]
    RepositoryPassCmd: str

    @field_validator("FilesAndFolders", mode="before")
    def expand_tilde_in_sources(cls, files_and_folders) -> set[str]:
        new = {str(Path(src).expanduser()) for src in files_and_folders}
        return new


Configuration = BtrFSRsyncConfig | ResticConfig


def parse_configuration(content: str) -> list[Configuration]:
    ConfigList = TypeAdapter(list[Configuration])
    config_lst = ConfigList.validate_json(content)
    if len(config_lst) == 0:
        sys.exit("Leere Konfigurationsdateien sind nicht erlaubt.\n")
    return config_lst
